# coding:utf-8

from .exceptions import UnacceptableResultError
from .queryformat import qformat
from .restructables import (RestructableList as RL,
                            RestructableDict as RD)


class Client(object):
    def __init__(self, dbapi, **kwargs):
        self.query_log = False

        self._dbapi = dbapi
        self._config = kwargs

        self._connection = None
        self._cursor = None

        self._result_cols = None
        self._result_rows = None
        self._lastrowid = None
        self._rowcount = None

    def __enter__(self):
        return self

    def __exit__(self, typ, val, tb):
        self.close()

    def connection(self):
        if self._connection is None:
            self._connection = self._dbapi.connect(**self._config)
        return self._connection

    def cursor(self):
        if self._cursor is None:
            self._cursor = self.connection().cursor()
        return self._cursor

    def commit(self, close=True):
        self.connection().commit()

        if close:
            self.close()

        return self

    def rollback(self, close=True):
        self.connection().rollback()

        if close:
            self.close()

        return self

    def close(self):
        if self._cursor is not None:
            self._cursor.close()
            self._cursor = None

        if self._connection is not None:
            self._connection.close()
            self._connection = None

        return self

    def exes(self, fmt, **kwargs):
        return self.exe(True, fmt, **kwargs)

    def exew(self, fmt, **kwargs):
        return self.exe(False, fmt, **kwargs)

    def exe(self, strong, fmt, **kwargs):
        q = qformat(fmt, **kwargs)

        if self.query_log:
            print(q)

        try:
            self.cursor().execute(q)

        except Exception as e:
            self.close()
            raise e

        else:
            self._lastrowid = self.cursor().lastrowid
            self._rowcount = self.cursor().rowcount

            self._result_cols = [c[0] for c in self.cursor().description or []]
            self._result_rows = [r for r in self.cursor()]

            if not strong:
                self.close()

            return self

    def scalar(self, allow_none=True):
        if len(self._result_rows) > 0:
            return self._result_rows[0][0]

        else:
            return None if allow_none else self._unacceptable_result()

    def lastid(self, allow_zero=True):
        if allow_zero:
            return self._lastrowid or 0

        else:
            return self._lastrowid or self._unacceptable_result()

    def rowcount(self, allow_zero=True):
        if allow_zero:
            return self._rowcount or 0

        else:
            return self._rowcount or self._unacceptable_result()

    def first(self, allow_none=True):
        if len(self._result_rows) > 0:
            return RD(dict(zip(self._result_cols, self._result_rows[0])))

        else:
            return None if allow_none else self._unacceptable_result()

    def all(self, allow_empty=True):
        r = RL([
            RD(dict(zip(self._result_cols, r))) for r in self._result_rows
        ])

        if 0 < len(r):
            return r

        else:
            return [] if allow_empty else self._unacceptable_result()

    def _unacceptable_result(self):
        raise UnacceptableResultError()



# coding:utf-8

import re
from datetime import datetime, date
from . import compat
from .exceptions import (QueryFormatError,
                         IncompatibleTypeError)

PLACEHOLDER = re.compile(r'([^:]*)(:+)([0-9a-zA-Z_]+)')


def qformat(fmt, **kwargs):
    def _(m):
        pre, head, key = m.groups()

        if pre == '\\':
            return head + key

        else:
            v = kwargs[key]
            should_escape = head == ':'
            return pre + qformat_object(v, should_escape)

    try:
        return PLACEHOLDER.sub(_, fmt)

    except KeyError as e:
        raise QueryFormatError(str(e))


def qformat_object_default(v, escape=True, nested=False):
    if isinstance(v, bool):
        return u'1' if v else u'0'

    if compat.is_number(v):
        return compat.unicode_str(v)

    if compat.is_non_unicode_string(v):
        return qformat_object(v.decode('utf8'), escape, nested)

    if compat.is_unicode_string(v):
        return u"'{}'".format(v.replace(u"'", u"\\'")) if escape else v

    if v is None:
        return u'NULL'

    if isinstance(v, (list, tuple, set)):
        s = u','.join([qformat_object(e, escape, True) for e in v])
        return u'({})'.format(s) if nested else s

    if isinstance(v, dict):
        d = {qformat_object(k, False, True): qformat_object(e, True, True)
             for k, e in v.items()}
        return u','.join([u'{}={}'.format(k, e) for k, e in d.items()])

    if hasattr(v, '__call__'):
        r = v()
        fname, args = (r[0], r[1:]) if isinstance(r, tuple) else (r, ())
        return u'{}({})'.format(fname.upper(), qformat_object(args))

    if isinstance(v, datetime):
        return u"'{}'".format(v.strftime('%Y-%m-%d %H:%M:%S'))

    if isinstance(v, date):
        dt = datetime(v.year, v.month, v.day)
        return "'{}'".format(dt.strftime('%Y-%m-%d'))

    raise IncompatibleTypeError(v, escape, nested)


qformat_object = qformat_object_default



# coding:utf-8

import sys
major, minor, micro, releaselevel, serial = sys.version_info


if major == 2:
    def is_number(v):
        return isinstance(v, (int, long, float))

    def is_unicode_string(v):
        return isinstance(v, unicode)

    def is_non_unicode_string(v):
        return isinstance(v, (str, bytes, bytearray))

    def unicode_str(v):
        return unicode(v)

    from itertools import izip_longest
    zip_longest = izip_longest

    def keypath_value(dic, *keypath):
        return reduce(lambda d, k: d.get(k), [dic] + list(keypath))

elif major == 3:
    def is_number(v):
        return isinstance(v, (int, float))

    def is_unicode_string(v):
        return isinstance(v, str)

    def is_non_unicode_string(v):
        return isinstance(v, (bytes, bytearray))

    def unicode_str(v):
        return str(v)

    from itertools import zip_longest

    from functools import reduce

    def keypath_value(dic, *keypath):
        return reduce(lambda d, k: d.get(k), [dic] + list(keypath))



# coding:utf-8

from . import compat


class RestructableDict(dict):
    def nest(self, *keypath):
        head, tail = keypath[0], keypath[1:]

        if len(keypath) == 1:
            keys = list(self.keys())
            d = {k[len(head)+1:]: self.pop(k) for k in keys
                 if k.startswith(head + '_')}
            self[head] = RestructableDict(d)

        else:
            self[head].nest(*tail)

        return self

    def split(self, *keypath):
        head, tail = keypath[0], keypath[1:]

        if len(keypath) == 1:
            v = self[head].split(',') if self[head] is not None else []
            self[head] = RestructableList(v)

        else:
            self[head].split(*tail)

        return self

    def invert(self, *keypath):
        head, tail = keypath[0], keypath[1:]

        if len(keypath) == 1:
            keys = self[head].keys()
            ls = [RestructableDict(dict(zip(keys, e)))
                  for e in compat.zip_longest(*self[head].values())]
            self[head] = RestructableList(ls)

        else:
            self[head].invert(*tail)

        return self

    def nest_list(self, *keypath):
        self.nest(*keypath)
        rst = compat.keypath_value(self, *keypath)
        for k in rst.keys():
            rst.split(k)
        return self.invert(*keypath)


class RestructableList(list):
    def nest(self, *keypath):
        for e in self:
            e.nest(*keypath)
        return self

    def split(self, *keypath):
        for e in self:
            e.split(*keypath)
        return self

    def invert(self, *keypath):
        for e in self:
            e.invert(*keypath)
        return self

    def nest_list(self, *keypath):
        for e in self:
            e.nest_list(*keypath)
        return self
